// The MIT License

// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Author: Kermin Fleming kfleming@mit.edu

import Connectable::*;
import GetPut::*;
import ClientServer::*;
import FIFO::*;
import SpecialFIFOs::*;
import Vector::*;

//AWB includes
`include "asim/provides/low_level_platform_interface.bsh"
`include "asim/provides/soft_connections.bsh"
`include "asim/provides/common_services.bsh"

//Local includes
`include "asim/provides/audio_pipe_types.bsh"  //provides Audio Pipeline interface
`include "asim/provides/path_types.bsh"
`include "asim/provides/core.bsh"
`include "asim/provides/mixer.bsh"
`include "asim/provides/processor_library.bsh"
`include "asim/provides/fpga_components.bsh"
`include "asim/dict/VDEV_SCRATCH.bsh"

`include "asim/rrr/remote_client_stub_AUDIOCORERRR.bsh"
`include "asim/rrr/remote_server_stub_AUDIOCORERRR.bsh"

`define MAX_VOICES       2

module [CONNECTED_MODULE] mkConnectedApplication ();

 //  Core core <- mkCore(`VDEV_SCRATCH_MEMORYA);
 //  Core anotherCore <- mkCore(`VDEV_SCRATCH_MEMORYB);
   Vector#(`MAX_VOICES, Volume) channelVols = replicate(127);
   Mixer mixer <- mkMixer(`MAX_VOICES, channelVols); //should be max voices but 2 for now
   
   Reg#(int) cycle <- mkReg(0);
   Reg#(int) sampleCount <-mkReg(0);
   Vector#(2, Reg#(Bool)) ac_fini <- replicateM(mkReg(False));

   // Services Samples 
   ClientStub_AUDIOCORERRR client_stub <- mkClientStub_AUDIOCORERRR();   
  

   //-----------------------------------------------------------
   // Debug port
   
   ServerStub_AUDIOCORERRR server_stub <- mkServerStub_AUDIOCORERRR();   

   // Create Cores
   Vector#(`MAX_VOICES, Core) cores;
   for (Integer n = 0; n < `MAX_VOICES; n = n + 1)
     begin
	 case (n)
	    0 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYA);
	    1 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYB);
	 endcase
     end
   
  // this is for the tracing
  rule printCycles;
    cycle <= cycle+1;
    $fdisplay(stderr, " => Cycle = %d", cycle);
  endrule

  // Send to Mixer
  // Right now this is sorta retarded in that I pass from the output fifo into a new fifo
  // But I have to mod a bunch of things to fix this and I am not sure I understand
  // things well enough to do this quickly.  So here it is as it is for now.
   rule mix;
      for (Integer i = 0; i < `MAX_VOICES; i = i + 1)
        begin
	   let coreOut <- cores[i].sampleOutput.get();
	   mixer.toMixer(AudioStream {voice : fromInteger(i), data: tagged Valid coreOut});
	end
      
//     let coreOut <- core.sampleOutput.get();
//    mixer.toMixer(AudioStream {voice : 0, data : tagged Valid coreOut});
//    mixer.toMixer(AudioStream {voice : 1, data : tagged Valid anotherOut});
  endrule

  rule feedOutput;
     let pipeOut <- mixer.mainOut.get();
     
     AudioProcessorControl endOfFileTag = EndOfFile;
     AudioProcessorControl sampleTag = Data;

     sampleCount <= sampleCount+1;
     
     $display("PIPE writes sample %d", sampleCount);
     case (pipeOut) matches
	tagged EndOfFile:
	   begin
	      client_stub.makeRequest_SendProcessedStream(zeroExtend(pack(endOfFileTag)),?);
	      ac_fini[0] <= True;
	   end
       tagged Sample .sample:
	  client_stub.makeRequest_SendProcessedStream(zeroExtend(pack(sampleTag)), zeroExtend(pack(sample)));
     endcase
  endrule

   // Can generally just stick with the EOF but since I have two Cores no mixer...
  /*(* conservative_implicit_conditions *)
   rule sendTerminate;
      Bool done = True;
      for (Integer i = 1; i < 2; i = i+1)
	 done = ac_fini[i] && done;

      if (done)
	 client_stub.makeRequest_SendTerminate(zeroExtend(pack(1)));
      
   endrule
*/

   //*****  SERVER Side  *****

   (* conservative_implicit_conditions *)
   rule feedInput;
     let command <- server_stub.acceptRequest_SendUnprocessedStream();
     AudioProcessorControl ctrl = unpack(truncate(command.ctrl));

     VoiceId channel = unpack(truncate(command.channel));
//      $display("rlm: %x", test);
	   
     AudioProcessorUnit inSample;
      
     if(ctrl == EndOfFile)
       begin
	   $display("lsp: PIPE received EOF   ");
	   inSample = tagged EndOfFile;
//           core.sampleInput.put(tagged EndOfFile);
       end
     else 
	begin
//	   $display("lsp: PIPE received Data   ");
//           core.sampleInput.put(tagged Sample unpack(truncate(command.sample)));	   
           inSample = tagged Sample unpack(truncate(command.sample));
	end

      cores[channel].sampleInput.put(inSample);
      
   endrule     
endmodule
